@@ -22,11 +22,11 @@ from snowflake.connector.options import pandas
 from snowflake.connector.telemetry import TelemetryData, TelemetryField
 
 from ._utils import (
-    _PYTHON_SNOWPARK_USE_SCOPED_TEMP_OBJECTS_STRING,
     TempObjectType,
     get_temp_type_for_object,
     random_name_for_temp_object,
 )
+from .constants import _PARAM_USE_SCOPED_TEMP_FOR_PANDAS_TOOLS
 from .cursor import SnowflakeCursor
 
 if TYPE_CHECKING:  # pragma: no cover
@@ -254,6 +254,7 @@ def write_pandas(
     on_error: str = "abort_statement",
     parallel: int = 4,
     quote_identifiers: bool = True,
+    infer_schema: bool = False,
     auto_create_table: bool = False,
     create_temp_table: bool = False,
     overwrite: bool = False,
@@ -261,6 +262,7 @@ def write_pandas(
     use_logical_type: bool | None = None,
     iceberg_config: dict[str, str] | None = None,
     bulk_upload_chunks: bool = False,
+    use_vectorized_scanner: bool = False,
     **kwargs: Any,
 ) -> tuple[
     bool,
@@ -308,11 +310,15 @@ def write_pandas(
         on_error: Action to take when COPY INTO statements fail, default follows documentation at:
             https://docs.snowflake.com/en/sql-reference/sql/copy-into-table.html#copy-options-copyoptions
             (Default value = 'abort_statement').
+        use_vectorized_scanner: Boolean that specifies whether to use a vectorized scanner for loading Parquet files. See details at
+            `copy options <https://docs.snowflake.com/en/sql-reference/sql/copy-into-table.html#copy-options-copyoptions>`_.
         parallel: Number of threads to be used when uploading chunks, default follows documentation at:
             https://docs.snowflake.com/en/sql-reference/sql/put.html#optional-parameters (Default value = 4).
         quote_identifiers: By default, identifiers, specifically database, schema, table and column names
             (from df.columns) will be quoted. If set to False, identifiers are passed on to Snowflake without quoting.
             I.e. identifiers will be coerced to uppercase by Snowflake.  (Default value = True)
+        infer_schema: Perform explicit schema inference on the data in the DataFrame and use the inferred data types
+            when selecting columns from the DataFrame. (Default value = False)
         auto_create_table: When true, will automatically create a table with corresponding columns for each column in
             the passed in DataFrame. The table will not be created if it already exists
         create_temp_table: (Deprecated) Will make the auto-created table as a temporary table
@@ -353,20 +359,13 @@ def write_pandas(
             f"Invalid compression '{compression}', only acceptable values are: {compression_map.keys()}"
         )
 
+    # TODO(SNOW-1505026): Get rid of this when the BCR to always create scoped temp for intermediate results is done.
     _use_scoped_temp_object = (
-        conn._session_parameters.get(
-            _PYTHON_SNOWPARK_USE_SCOPED_TEMP_OBJECTS_STRING, False
-        )
+        conn._session_parameters.get(_PARAM_USE_SCOPED_TEMP_FOR_PANDAS_TOOLS, False)
         if conn._session_parameters
         else False
     )
 
-    """sfc-gh-yixie: scoped temp stage isn't required out side of a SP.
-    TODO: remove the following line when merging SP connector and Python Connector.
-    Make sure `create scoped temp stage` is supported when it's not run in a SP.
-    """
-    _use_scoped_temp_object = False
-
     if create_temp_table:
         warnings.warn(
             "create_temp_table is deprecated, we still respect this parameter when it is True but "
@@ -486,7 +485,7 @@ def write_pandas(
             num_statements=1,
         )
 
-    if auto_create_table or overwrite:
+    if auto_create_table or overwrite or infer_schema:
         file_format_location = _create_temp_file_format(
             cursor,
             database,
@@ -529,27 +528,29 @@ def write_pandas(
             quote_identifiers,
         )
 
-        iceberg = "ICEBERG " if iceberg_config else ""
-        iceberg_config_statement = _iceberg_config_statement_helper(
-            iceberg_config or {}
-        )
+        if auto_create_table or overwrite:
+            iceberg = "ICEBERG " if iceberg_config else ""
+            iceberg_config_statement = _iceberg_config_statement_helper(
+                iceberg_config or {}
+            )
+
+            create_table_sql = (
+                f"CREATE {table_type.upper()} {iceberg}TABLE IF NOT EXISTS identifier(?) "
+                f"({create_table_columns}) {iceberg_config_statement}"
+                f" /* Python:snowflake.connector.pandas_tools.write_pandas() */ "
+            )
+            params = (target_table_location,)
+            logger.debug(
+                f"auto creating table with '{create_table_sql}'. params: %s", params
+            )
+            cursor.execute(
+                create_table_sql,
+                _is_internal=True,
+                _force_qmark_paramstyle=True,
+                params=params,
+                num_statements=1,
+            )
 
-        create_table_sql = (
-            f"CREATE {table_type.upper()} {iceberg}TABLE IF NOT EXISTS identifier(?) "
-            f"({create_table_columns}) {iceberg_config_statement}"
-            f" /* Python:snowflake.connector.pandas_tools.write_pandas() */ "
-        )
-        params = (target_table_location,)
-        logger.debug(
-            f"auto creating table with '{create_table_sql}'. params: %s", params
-        )
-        cursor.execute(
-            create_table_sql,
-            _is_internal=True,
-            _force_qmark_paramstyle=True,
-            params=params,
-            num_statements=1,
-        )
         # need explicit casting when the underlying table schema is inferred
         parquet_columns = "$1:" + ",$1:".join(
             f"{quote}{snowflake_col}{quote}::{column_type_mapping[col]}"
@@ -586,8 +587,9 @@ def write_pandas(
             f"FROM (SELECT {parquet_columns} FROM '{copy_stage_location}') "
             f"FILE_FORMAT=("
             f"TYPE=PARQUET "
+            f"USE_VECTORIZED_SCANNER={use_vectorized_scanner} "
             f"COMPRESSION={compression_map[compression]}"
-            f"{' BINARY_AS_TEXT=FALSE' if auto_create_table or overwrite else ''}"
+            f"{' BINARY_AS_TEXT=FALSE' if auto_create_table or overwrite or infer_schema else ''}"
             f"{sql_use_logical_type}"
             f") "
             f"PURGE=TRUE ON_ERROR=?"
